import { webcrypto, createHmac, createDecipheriv, createHash } from "crypto"
import { HMACAssertionError } from "./errors"

declare module "crypto" {
  export const webcrypto: Crypto
}

/** Encryption and MAC */
export interface Cipher {
  /** Encryption key */
  key: Buffer
  /** HMAC key */
  hmac: Buffer
}

async function pbkdf2(password: string, salt: string, iterations = 1000, length = 256) {
  const encoder = new TextEncoder()
  const key = await webcrypto.subtle.importKey(
    "raw",
    encoder.encode(password),
    "PBKDF2",
    false,
    ["deriveBits"]
  )
  const bits = await webcrypto.subtle.deriveBits(
    {
      name: "PBKDF2",
      hash: "SHA-512",
      salt: encoder.encode(salt),
      iterations,
    },
    key,
    length
  )
  return bits
}

export function assertHMac(data: Buffer, key: Buffer, expected: Buffer) {
  const actual = createHmac("sha256", key).update(data).digest()
  if (!actual.equals(expected)) {
    throw new HMACAssertionError()
  }
}

export const splitPlainText = (derivedKey: Buffer): Cipher => ({
  key: derivedKey.slice(0, 32),
  hmac: derivedKey.slice(32, 64),
})

export function decryptKeys(encryptedKey: string, derived: Cipher) {
  const buffer = Buffer.from(encryptedKey, "base64")
  const base = decryptOPData(buffer, derived)
  const digest = createHash("sha512").update(base).digest()
  return splitPlainText(digest)
}

export function decryptData(key: Buffer, iv: Buffer, data: Buffer) {
  const cipher = createDecipheriv("aes-256-cbc", key, iv).setAutoPadding(false)
  return Buffer.concat([cipher.update(data), cipher.final()])
}

export function decryptOPData(cipherText: Buffer, cipher: Cipher) {
  const key = cipherText.slice(0, -32)
  assertHMac(key, cipher.hmac, cipherText.slice(-32))

  const plaintext = decryptData(cipher.key, key.slice(16, 32), key.slice(32))
  const size = readUint16(key.slice(8, 16))
  return plaintext.slice(-size)
}

function readUint16({ buffer, byteOffset, length }: Buffer) {
  return new DataView(buffer, byteOffset, length).getUint16(0, true)
}
