import { assertHMac, decryptData, decryptOPData, splitPlainText } from "../crypto"
import type { ItemDetails } from "../types"
import type { Cipher } from "../crypto"
import type { Attachment } from "./attachment"

export interface EncryptedItem {
  category: string // "001"
  /** Unix seconds */
  created: integer
  d: string // "b3BkYXRhMbt"
  folder: string // 32 chars
  hmac: string // base64
  k: string // base64
  o: string // base64
  tx: integer // Unix seconds
  updated: integer // Unix seconds
  uuid: string // 32 chars
}

export class ItemAttachmentBridge {
  constructor(public cipher?: Cipher) {}
}

export class Item implements Disposable {
  #data: EncryptedItem

  itemDetails!: ItemDetails
  attachments: Attachment[] = []

  get uuid() {
    return this.#data.uuid
  }

  constructor(data: EncryptedItem) {
    this.#data = data
  }

  dispose() {
    this.attachments.forEach(a => a.dispose())
    this.itemDetails = null!
  }

  #deriveConcreteKey(master: Cipher) {
    const k = Buffer.from(this.#data.k, "base64")
    const data = k.slice(0, -32)
    assertHMac(data, master.hmac, k.slice(-32))
    const derivedKey = decryptData(master.key, data.slice(0, 16), data.slice(16))
    return splitPlainText(derivedKey)
  }

  /** @internal */
  decryptItemDetail(master: Cipher): ItemDetails {
    const cipher = this.#deriveConcreteKey(master)
    const detail = decryptOPData(
      /* cipherText */ Buffer.from(this.#data.d, "base64"),
      /* cipher */ cipher
    )
    this.itemDetails = JSON.parse(detail.toString("utf-8"))
    return this.itemDetails
  }

  /** @internal */
  decryptAttachments(master: Cipher) {
    const cipher = this.#deriveConcreteKey(master)
    this.attachments.forEach(a => a.decrypt(cipher))
  }
}
