import * as crypto from "crypto"
import type { IAdapter } from "../adapters"
import type { Cipher } from "../crypto"
import { decryptKeys, splitPlainText, decryptOPData } from "../crypto"
import { HMACAssertionError, invariant } from "../errors"
import { OnePasswordFileManager } from "../fs"
import { i18n } from "../i18n"
import type { EncryptedItem } from "./item"
import { Item } from "./item"
import type { Profile, Overview } from "../types"
import { WeakValueMap } from "../weakMap"
import { EventEmitter } from "../ee"

type Band = { [uuid: string]: Item }
type FoldersMap = { [uuid: string]: Band }

interface VaultEvents {
  lock: void
}

/**
 * Main OnePassword Vault class
 */
export class Vault extends EventEmitter<VaultEvents> {
  // Ciphers
  #master?: Cipher
  #overview?: Cipher

  // File system interface
  #files: OnePasswordFileManager

  #profile: Profile
  #folders: FoldersMap

  #overviews = new Map<string, Overview>()
  #items: Item[] = []
  #itemsMap = new WeakValueMap<string, Item>()

  private constructor(
    files: OnePasswordFileManager,
    profile: Profile,
    folders: FoldersMap,
    items: Item[]
  ) {
    super()
    this.#files = files
    this.#profile = profile
    this.#folders = folders
    this.#items = items

    items.forEach(item => {
      this.#itemsMap.set(item.uuid, item)
    })
  }

  /**
   * Create a new OnePassword Vault instance and read all bands.
   * @internal
   */
  static async of(path: string, profileName = "default", adapter: IAdapter) {
    const files = OnePasswordFileManager(adapter.fs, path, profileName)
    const profile = JSON.parse(
      stripText(await files.getProfile(), /^var profile\s*=/, ";")
    )
    const folders = JSON.parse(stripText(await files.getFolders(), "loadFolders(", ");"))
    const bands: Item[] = []

    for (let i = 0; i < 36; i++) {
      const letter = i.toString(36).toUpperCase()
      const source = await files.getBand(letter)
      if (!source) continue
      const object = JSON.parse(stripText(source, "ld(", ");"))
      for (const value of Object.values(object)) {
        bands.push(new Item(value as EncryptedItem))
      }
    }

    return new Vault(files, profile, folders, bands)
  }

  readonly overviews = Object.freeze({
    get: (condition: string | ((overview: Overview) => boolean)) => {
      this.#assertUnlocked()

      if (typeof condition === "string") {
        const title = condition
        condition = overview => overview.title === title
      }
      for (const value of this.#overviews.values()) {
        if (condition(value)) {
          return value
        }
      }
    },

    values: () => {
      this.#assertUnlocked()
      return Array.from(this.#overviews.values())
    },
  })

  /**
   * Unlock this OnePassword vault.
   * @param masterPassword User provided master password. Only the derived
   * master and overview key will be stored within the class.
   */
  unlock(masterPassword: string) {
    const profile = this.#profile
    const derivedKey = crypto.pbkdf2Sync(
      /* password */ masterPassword,
      /* salt */ toBuffer(profile.salt),
      /* iterations */ profile.iterations,
      /* keylen */ 64,
      /* digest */ "sha512"
    )

    const cipher = splitPlainText(derivedKey)

    // Derive master key and overview keys
    try {
      this.#master = decryptKeys(profile.masterKey, cipher)
      this.#overview = decryptKeys(profile.overviewKey, cipher)
    } catch (e) {
      if (e instanceof HMACAssertionError) {
        throw new Error(i18n.error.invalidPassword)
      }
      throw e
    }

    this.#readOverviews()
    return this
  }

  /**
   * Remove derived keys stored within the class instance.
   */
  lock() {
    this.#master = null!
    this.#overview = null!
    this.#overviews.clear()
    this.emit("lock")
    return this
  }

  get isLocked() {
    return this.#master?.key == null
  }

  #assertUnlocked() {
    if (this.isLocked) {
      throw new Error(i18n.error.vaultIsLocked)
    }
  }

  getItem(uuid: string) {
    this.#assertUnlocked()
    const item = this.#itemsMap.get(uuid)
    if (!item) return

    if (!item.itemDetails) {
      item.decryptItemDetail(this.#master!)
    }

    const encrypted = uuid ? this.#encryptedItems.get(uuid[0])![uuid] : undefined
    if (!encrypted) return

    return {
      original: encrypted,
      details: decryptItemDetail(encrypted, this.#master!),
    }
  }

  #readOverviews() {
    this.#assertUnlocked()
    this.#encryptedItems.forEach(value => {
      for (const [uuid, item] of Object.entries(value)) {
        const overview = decryptOverview(item, this.#overview!)
        overview.uuid = uuid
        this.#overviews.set(uuid, overview)
      }
    })
  }
}

function decryptOverview(item: EncryptedItem, overviewCipher: Cipher): Overview {
  try {
    const overview = decryptOPData(toBuffer(item.o), overviewCipher)
    return JSON.parse(overview.toString("utf8")) as Overview
  } catch (e) {
    console.error(i18n.error.cannotDecryptOverviewItem)
    throw e
  }
}

function stripText(text: string, prefix: string | RegExp, suffix: string | RegExp) {
  if (typeof prefix === "string") {
    if (text.startsWith(prefix)) {
      text = text.slice(prefix.length)
    }
  } else {
    const prefixMatch = text.match(prefix)
    if (prefixMatch) {
      text = text.slice(prefixMatch[0].length)
    }
  }
  if (typeof suffix === "string") {
    if (text.endsWith(suffix)) {
      text = text.slice(0, -suffix.length)
    }
  } else {
    const suffixMatch = text.match(suffix)
    if (suffixMatch) {
      text = text.slice(0, -suffixMatch[0].length)
    }
  }
  return text
}

function toBuffer(data: string) {
  return Buffer.from(data, "base64")
}
