import type { FieldType } from "./models"

type integer = number

export interface Profile {
  lastUpdatedBy: "Dropbox"
  /** Unix seconds */
  updatedAt: integer
  profileName: string
  salt: string // base64
  masterKey: string // base64
  iterations: integer // 50000
  uuid: string // 32 chars
  overviewKey: string // "b3B...IMO52D"
  createdAt: integer // Unix seconds
}

export type TextField = {
  type: FieldType.Text
  value: string
  designation: string
  name: string
}
export type BooleanField = {
  type: FieldType.Checkbox
  name: string
  value?: "✓" | string
}

export type ItemField =
  | TextField
  | BooleanField
  | {
      // @TODO: This currently catches all item fields.
      type: FieldType
      value: string
      designation?: string
      name: string
    }

declare namespace ItemSection {
  type A = {
    guarded: "yes"
    clipboardFilter?: string
  }

  type String = {
    k: "string"
    v: string
    /** Unique name */
    n: string // "firstname" | "initial" | "address" | "class" | "conditions" | "expiry_date"
    a?: A
    /** User-readable title */
    t: string // "first name" | "initial" | "address" | "license class" | "conditions / restrictions" | "expiry date"
  }
  type Menu = {
    k: "menu"
    v: string // "female"
    a: A
    t: string // "sex"
  }
  type Date = {
    k: "date"
    v: number // 359100000
    n: string // "birthdate"
    a: A
    t: string // "birth date" | "date of birth"
  }
  type Gender = {
    k: "gender"
    n: "sex"
    v: string // "female"
    t: "sex"
  }
  type MonthYear = {
    k: "monthYear"
    n: string // "expiry_date"
    v: number // 2515
    t: string // "expiry date"
  }
  type Concealed = {
    k: "concealed"
    n: "password"
    v: string
    a?: {
      generate: "off"
    }
    t: "password"
  }

  // eslint-disable-next-line @typescript-eslint/ban-types
  type Any = String | Menu | Date | Gender | MonthYear | Concealed
}

// One of them is empty?, 0C4F27910A64488BB339AED63565D148
export interface ItemDetails {
  htmlForm?: {
    htmlAction: string // "/login/"
    htmlMethod: "post" | "get"
  }
  notesPlain?: string
  sections: {
    name: string // "name" | "title" | "internet"
    title: string // "Identification" | "Address", "Internet Details"
    fields?: ItemSection.Any[]
  }[]
  /** Web form fields */
  fields?: ItemField[]
}

export interface Folder {
  created: number // 1373754128
  overview: string // "b3BkYXRhT/../KBM="
  smart?: true
  tx: number // 1373754523
  /** Updated time, UNIX seconds */
  updated: number
  uuid: string // "AC78552EB06A4F65BEBF58B4D9E32080"
}

export interface Overview {
  /**
   * Password strength.
   *
   * `ps is the value for manually entered passwords and is used to display the relative strength value.`
   */
  ps: number

  /**
   * Password Bits of Entropy used for generating the password.
   */
  pbe?: number

  /**
   * https://1password.community/discussion/comment/372147/#Comment_372147
   * pbe should be used for password strength display
   */
  pgrng?: boolean
  title?: string
  ainfo?: string
  url?: string
  tags?: string[]
  URLs?: { u: string }[]
  uuid?: string // Added manually
  appIds: {
    name: string // 'Firefox' | 'Email'
    type: string // 'accessibility://login'
    id: string // 'android://...'
  }[]
}
